# Autogenerated by Astropy's setup.py on 2019-02-24 01:00:23 UTC
from __future__ import unicode_literals
import datetime

version = "2.0.12"
githash = "e2d0f0276d484fa4852b0a16c4944fa1076f0fbd"


major = 2
minor = 0
bugfix = 12

release = True
timestamp = datetime.datetime(2019, 2, 24, 1, 0, 23)
debug = False

astropy_helpers_version = "3.1"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
